(ns bombnet.game
  (:require [clojure.test :refer [is with-test]])
  (:require [clojure.set :refer :all]))

(defn new-board [w h]
  (let [should-wall (fn [row col]
                      (or
                        (= row 0)
                        (= row (- h 1))
                        (= col 0)
                        (= col (- w 1))
                        (= 0 (mod row 2) (mod col 2))))]
    (vec (map (fn [row]
                (vec (map (fn [col]
                            (if (should-wall row col) "#" " "))
                          (range 0 w))))
              (range 0 h )))))

(with-test
  (defn get-pos [entity]
    (cond
      (vector? entity) entity
      (:pos entity) (:pos entity)))
  (is [2 3] (get-pos [2 3]))
  (is [2 3] (get-pos {:pos [2 3]})))

(defn get-cell [board entity]
  (let [[x y] (get-pos entity)] (get-in board [y x])))

(defn new-player [id] {:id id :pos [1 1] :color "red"})

(defn move-player
  ([p x y] (assoc p :pos [x y]))
  ([p dir]
   (let [[x y] (:pos p)]
     (condp = dir
       "up" (move-player p x (- y 1))
       "down" (move-player p x (+ y 1))
       "left" (move-player p (- x 1) y)
       "right" (move-player p (+ x 1) y)))))

(defn new-game []
  {:board (new-board 21 17)
   :players []
   :bombs []})

(defn with-action [typ]
  (fn [p] (= typ (get-in p [:action :type]))))

(defn check-positions [state]
  "Returns a vector of valid new player positions"
  (let [{:keys [board players bombs]} state
        wall-clear (fn [p] (= " " (get-cell board p)))
        collides (fn [a b] (and (not= a b) (= (get-pos a) (get-pos b))))
        swap-clear (fn [p] (every? (fn [e]
                                     (not (and
                                            (= (:prev-pos p) (:pos e))
                                            (= (:prev-pos e) (:pos p)))))
                                   players))
        entity-clear (fn [entities] (fn [p] (every? #(not (collides p %)) entities)))
        xf (comp
             (filter wall-clear)
             (filter (entity-clear players))
             (filter (entity-clear bombs))
             (filter swap-clear)
             (map :id))]
    (into #{} xf players)))

(defn perform-movement [initial-state]
  (loop [state initial-state]
    (let [{players :players} state
          wants-to-move (with-action "move")
          to-move (vec (map (fn [p]
                              (let [p2 (assoc p :prev-pos (get p :pos))]
                                (if (wants-to-move p2)
                                  (dissoc (move-player p2 (get-in p2 [:action :dir])) :action)
                                  p)))
                            players))
          proposed-state (assoc state :players to-move)
          valid-moves (check-positions proposed-state)
          valid-players (map (fn [p]
                               (if (and
                                     (not (valid-moves (:id p)))
                                     (wants-to-move p))
                                 (dissoc p :action)
                                 p)) players)]
      (if (= players valid-players)
        proposed-state
        (recur (assoc state :players valid-players))))))

(defn ^:private offset [pos dir] (mapv + pos dir))

(with-test
  (defn explode-bomb-helper [board start dir power acc]
    (let [pos (offset start dir)]
      (if (or (zero? power) (= (get-cell board pos) "#"))
        acc
        (recur board pos dir (- power 1) (conj acc pos)))))
  (let [board (new-board 7 7)]
    (is (= #{[1 2] [1 3]}
           (explode-bomb-helper board [1 1] [0 1] 2 #{})))
    (is (= #{[1 2] [1 3] [1 4] [1 5]}
           (explode-bomb-helper board [1 1] [0 1] 99 #{})))))

(defn explode-bomb [board bomb]
  (let [{:keys [power pos]} bomb
        dirs (if (:diagonal? bomb)
               [[1 1] [-1 -1] [1 -1] [-1 1]]
               [[1 0] [0 1] [-1 0] [0 -1]])]
    (reduce #(explode-bomb-helper board pos %2 power %) #{pos} dirs)))

(defn explode-bombs
  ([state] (explode-bombs state #{}))
  ([state prev-explosion]
   (if-let [exploding (seq (filter #(zero? (:counter %)) (:bombs state)))]
     (let [{:keys [board players bombs]} state
           bomb (first exploding)
           other-bombs (filter #(not= bomb %) bombs)
           explosion (explode-bomb board bomb)
           new-bombs (mapv (fn [bomb]
                            (if (explosion (:pos bomb))
                              (assoc bomb :counter 0)
                              bomb))
                          other-bombs)
           new-players (mapv (fn [player]
                              (if (explosion (:pos player))
                                (assoc player :dead? true)
                                player))
                            players)]
       (recur
         (-> state
             (assoc :players new-players)
             (assoc :bombs new-bombs))
         (union prev-explosion explosion)))
     (assoc state :explosion prev-explosion))))
